﻿using System;
using System.Collections.Generic;
using GameUtils.Graphics;
using GameUtils.Graphics.Data;
using GameUtils.Math;
using SharpDX;
using SharpDX.Direct2D1;
using Vector2 = GameUtils.Math.Vector2;

namespace GameUtils.Renderers.Direct2D1
{
    internal class Direct2DGeometry : GeometryData
    {
        PathGeometry geometry;
        GeometrySink sink;
        readonly SharpDX.Direct2D1.Factory factory;
        Vector2 lastPoint;
        bool isClosed;
        const float flatteningTolerance = 0.001f;

        internal Direct2DGeometry(SharpDX.Direct2D1.Factory factory)
        {
            this.factory = factory;
            geometry = new PathGeometry(factory);
            sink = geometry.Open();
            sink.SetFillMode(FillMode.Winding);
        }

        internal PathGeometry Geometry
        {
            get { return geometry; }
        }

        private void AddArcInner(Vector2 center, float width, float height, float theta, float etha1, float etha2)
        {
            
        }

        protected override void AddArc(Arc arc)
        {
            throw new NotImplementedException();
        }

        protected override void AddLine(Vector2 point)
        {
            sink.AddLine(new SharpDX.Vector2(point.X, point.Y));
            lastPoint = point;
        }

        protected override void AddLines(Vector2[] points)
        {
            if (points == null)
                throw new ArgumentNullException("points");

            var drawingPoints = new SharpDX.Vector2[points.Length];
            for (int i = 0; i < points.Length; i++)
                drawingPoints[i] = new SharpDX.Vector2(points[i].X, points[i].Y);
            sink.AddLines(drawingPoints);
            lastPoint = points[points.Length - 1];
        }

        protected override void BeginFigure(Vector2 point, Graphics.FigureBegin begin)
        {
            sink.BeginFigure(new SharpDX.Vector2(point.X, point.Y), (SharpDX.Direct2D1.FigureBegin)begin);
            lastPoint = point;
        }

        protected override void EndFigure(Graphics.FigureEnd end)
        {
            sink.EndFigure((SharpDX.Direct2D1.FigureEnd)end);
        }

        protected override void AddBezier(Bezier bezier)
        {
            var segment = new BezierSegment();
            segment.Point1 = new SharpDX.Vector2(bezier.ControlPoint1.X, bezier.ControlPoint1.Y);
            segment.Point2 = new SharpDX.Vector2(bezier.ControlPoint2.X, bezier.ControlPoint2.Y);
            segment.Point3 = new SharpDX.Vector2(bezier.EndPoint.X, bezier.EndPoint.Y);
            sink.AddBezier(segment);
            lastPoint = bezier.EndPoint;
        }

        protected override void AddBeziers(Bezier[] beziers)
        {
            if (beziers == null)
                throw new ArgumentNullException("beziers");

            var segments = new BezierSegment[beziers.Length];
            for (int i = 0; i < beziers.Length; i++)
            {
                segments[i].Point1 = new SharpDX.Vector2(beziers[i].ControlPoint1.X, beziers[i].ControlPoint1.Y);
                segments[i].Point2 = new SharpDX.Vector2(beziers[i].ControlPoint2.X, beziers[i].ControlPoint2.Y);
                segments[i].Point3 = new SharpDX.Vector2(beziers[i].EndPoint.X, beziers[i].EndPoint.Y);
            }
            sink.AddBeziers(segments);
            lastPoint = beziers[beziers.Length - 1].EndPoint;
        }

        protected override void AddCurve(Vector2[] points, float tension)
        {
            if (points == null)
                throw new ArgumentNullException("points");

            float s = (1 - tension) / 2;
            var pointList = new List<Vector2>();
            pointList.Add(lastPoint);
            pointList.AddRange(points);
            points = pointList.ToArray();
            var tangents = new Vector2[points.Length];
            for (int i = 1; i < points.Length - 1; i++)
                tangents[i] = s * (points[i + 1] - points[i - 1]);
            tangents[0] = s * (points[1] - points[0]);
            tangents[points.Length - 1] = s * (points[points.Length - 1] - points[points.Length - 2]);
            var segments = new BezierSegment[points.Length - 1];
            for (int i = 0; i < segments.Length; i++)
            {
                var segment = new BezierSegment();
                Vector2 control1 = points[i] + tangents[i] / 3;
                Vector2 control2 = points[i + 1] - tangents[i + 1] / 3;
                segment.Point1 = new SharpDX.Vector2(control1.X, control1.Y);
                segment.Point2 = new SharpDX.Vector2(control2.X, control2.Y);
                segment.Point3 = new SharpDX.Vector2(points[i + 1].X, points[i + 1].Y);
                segments[i] = segment;
            }
            sink.AddBeziers(segments);
            lastPoint = points[points.Length - 1];
        }

        protected override void AddGeometry(Graphics.Geometry geometry, GeometryCombineMode combineMode)
        {
            if (geometry == null)
                throw new ArgumentNullException("geometry");

            Direct2DGeometry d2dGeometry;
            try
            {
                d2dGeometry = (Direct2DGeometry)this.GetResourceData(geometry);
            }
            catch (InvalidCastException ex)
            {
                throw new ArgumentException("Dieses Geometry-Objekt ist nicht kompatibel.", ex);
            }

            PathGeometry other = d2dGeometry.Geometry;
                
            PathGeometry newGeometry = new PathGeometry(factory);
            GeometrySink newSink = newGeometry.Open();
            newSink.SetFillMode(FillMode.Winding);
            sink.Close();
            this.geometry.Combine(other, (CombineMode)combineMode, flatteningTolerance, newSink);
            sink.Dispose();
            this.geometry.Dispose();
            this.geometry = newGeometry;
            sink = newSink;
        }

        protected override void AddRectangle(Math.Rectangle rect)
        {
            var other = new RectangleGeometry(factory, Direct2DHelper.ConvertRectangle(rect));
            var newGeometry = new PathGeometry(factory);
            GeometrySink newSink = newGeometry.Open();
            newSink.SetFillMode(FillMode.Winding);
            sink.Close();
            geometry.Combine(other, CombineMode.Union, flatteningTolerance, newSink);
            other.Dispose();
            sink.Dispose();
            geometry.Dispose();
            geometry = newGeometry;
            sink = newSink;
        }

        protected override void AddEllipse(Math.Ellipse ellipse)
        {
            var other = new EllipseGeometry(factory, new SharpDX.Direct2D1.Ellipse(new SharpDX.Vector2(ellipse.Center.X, ellipse.Center.Y), ellipse.RadiusX, ellipse.RadiusY));
            var newGeometry = new PathGeometry(factory);
            GeometrySink newSink = newGeometry.Open();
            newSink.SetFillMode(FillMode.Winding);
            sink.Close();
            geometry.Combine(other, CombineMode.Union, flatteningTolerance, newSink);
            other.Dispose();
            sink.Dispose();
            geometry.Dispose();
            geometry = newGeometry;
            sink = newSink;
        }

        protected override void Transform(Matrix2x3 matrix)
        {
            if (!isClosed) sink.Close();
            var d2dMatrix = new Matrix3x2();
            d2dMatrix.TranslationVector = new SharpDX.Vector2(matrix.OffsetX, matrix.OffsetY);
            d2dMatrix.ScaleVector = new SharpDX.Vector2(matrix[0, 0], matrix[1, 1]);
            d2dMatrix.M12 = matrix[1, 0];
            d2dMatrix.M21 = matrix[0, 1];
            var transformed = new TransformedGeometry(factory, geometry, d2dMatrix);
            var dummy = new PathGeometry(factory);
            dummy.Open().Close();
            var newGeometry = new PathGeometry(factory);
            GeometrySink newSink = newGeometry.Open();
            newSink.SetFillMode(FillMode.Winding);
            transformed.Combine(dummy, CombineMode.Union, flatteningTolerance, newSink);
            dummy.Dispose();
            sink.Dispose();
            geometry.Dispose();
            geometry = newGeometry;
            sink = newSink;
            if (isClosed) sink.Close();
            lastPoint = matrix.ApplyTo(lastPoint);
        }

        protected override void Close()
        {
            sink.Close();
            isClosed = true;
        }

        protected override void Dispose(bool disposing)
        {
            sink.Dispose();
            geometry.Dispose();
        }
    }
}
