﻿using System;
using System.Collections.Generic;
using GameUtils.Math;

namespace GameUtils.Graphics
{
    public sealed class EngineIntro : RegistrationContext<EngineIntro.IntroState>, IDisposable
    {
        readonly List<IntroState> updateables; 
        readonly IntroRenderer renderer;
        bool eventRisen;

        public event EventHandler AnimationComplete;

        public sealed class IntroState : IBufferedState<IntroState>, IDisposable
        {
            float scaleSpeed;
            double elapsedMilliseconds;

            public event EventHandler AnimationComplete;

            public float GearScale { get; private set; }

            public float GearRotation { get; private set; }

            public SolidColorBrush GearBrush { get; private set; }

            public SolidColorBrush TextBrush { get; private set; }

            internal IntroState()
            {
                GearScale = 5.0f;
                scaleSpeed = 1.0f;
            }

            private void OnAnimationComplete(EventArgs e)
            {
                if (AnimationComplete != null)
                    AnimationComplete(this, e);
            }

            void IBufferedState<IntroState>.Update(IntroState oldState, TimeSpan elapsed)
            {
                elapsedMilliseconds = oldState.elapsedMilliseconds + elapsed.TotalMilliseconds;

                scaleSpeed = System.Math.Max(oldState.scaleSpeed - 0.125f * (float)elapsed.TotalSeconds, 0.0f);
                GearScale = System.Math.Max(oldState.GearScale - scaleSpeed * (float)elapsed.TotalSeconds, 1.0f);
                GearRotation = oldState.GearRotation + MathHelper.Pi * 0.1f * (float)elapsed.TotalSeconds;

                GearBrush = oldState.GearBrush;
                if (GearBrush == null)
                {
                    GearBrush = new SolidColorBrush(Color4.White);
                }
                else if (elapsedMilliseconds >= 10000)
                {
                    GearBrush.Opacity -= 0.5f * (float)elapsed.TotalSeconds;
                }

                TextBrush = oldState.TextBrush;
                if (TextBrush == null)
                {
                    TextBrush = new SolidColorBrush(Color4.White);
                    TextBrush.Opacity = 0;
                }
                else if (elapsedMilliseconds >= 10000)
                {
                    TextBrush.Opacity -= 0.5f * (float)elapsed.TotalSeconds;
                }
                else if (elapsedMilliseconds >= 5500)
                {
                    TextBrush.Opacity += 0.5f * (float)elapsed.TotalSeconds;
                }

                if (elapsedMilliseconds >= 15000)
                    OnAnimationComplete(EventArgs.Empty);
            }

            private bool disposed;

            public void Dispose()
            {
                Dispose(true);
                GC.SuppressFinalize(this);
            }

            private void Dispose(bool disposing)
            {
                if (!disposed)
                {
                    if (GearBrush != null) GearBrush.Dispose();
                    if (TextBrush != null) TextBrush.Dispose();

                    disposed = true;
                }
            }

            ~IntroState()
            {
                Dispose(false);
            }
        }

        public sealed class IntroRenderer : IStateRenderer<IntroState>, IDisposable
        {
            readonly Font font;
            readonly TextFormat format;
            readonly Geometry gear;
            readonly Geometry gearhole;

            public event EventHandler DepthPositionChanged;

            float IStateRenderer<IntroState>.DepthPosition
            {
                get { return float.MinValue; }
            }

            internal IntroRenderer()
            {
                font = new Font("Segoe UI", 12f);
                format = new TextFormat();
                gearhole = new Geometry();
                gear = new Geometry();

                format.HorizontalAlignment = HorizontalAlignment.Center;
                format.VerticalAlignment = VerticalAlignment.Center;

                gearhole.AddEllipse(new Ellipse(Vector2.Zero, 6f, 6f));
                gearhole.Close();

                const int cogCount = 8;
                const float piOverCogCount = MathHelper.Pi / (cogCount / 2);
                const float gearRadius = 20f;
                const float cogWidth = 2 * MathHelper.Pi * gearRadius / (2 * cogCount);
                const float cogDepth = 0.25f * gearRadius;
                for (int i = 0; i < cogCount; i++)
                {
                    var cogCenter = new Vector2(gearRadius * MathHelper.Cos(i * piOverCogCount), gearRadius * MathHelper.Sin(i * piOverCogCount));
                    var cogVector = cogCenter.CrossProduct().Normalize() * (cogWidth / 2);
                    var cogNormal = cogCenter.Normalize() * -cogDepth;
                    var outer1 = cogCenter + cogVector;
                    var inner1 = outer1 + cogNormal;
                    var outer2 = cogCenter - cogVector;
                    var inner2 = outer2 + cogNormal;

                    if (i == 0)
                        gear.BeginFigure(inner2, FigureBegin.Filled);
                    else
                        gear.AddLine(inner2);
                    gear.AddLine(outer2);
                    gear.AddLine(outer1);
                    gear.AddLine(inner1);
                }
                gear.EndFigure(FigureEnd.Closed);
                gear.AddEllipse(new Ellipse(Vector2.Zero, 16.5f, 16.5f));
                gear.CombineWith(gearhole, GeometryCombineMode.Xor);
                gear.Close();
            }

            void IStateRenderer<IntroState>.Render(IntroState state, Renderer renderer)
            {
                Matrix2x3 transform = Matrix2x3.Translation(renderer.SurfaceBounds.Width / 2.0f, renderer.SurfaceBounds.Height / 2.0f)
                                      * Matrix2x3.Scaling(System.Math.Min(renderer.SurfaceBounds.Width, renderer.SurfaceBounds.Height) / 100.0f);

                renderer.SetTransform(transform);
                renderer.DrawText("Powered by", font, state.TextBrush, new Rectangle(-100, -40f, 200, 12f), format);
                renderer.DrawText("GameUtils", font, state.TextBrush, new Rectangle(-100, 25f, 200, 12f), format);

                transform *= Matrix2x3.Scaling(state.GearScale) * Matrix2x3.Rotation(state.GearRotation);
                renderer.SetTransform(transform);
                renderer.FillGeometry(gear, state.GearBrush);

                renderer.ResetTransform();
            }

            private bool disposed;

            public void Dispose()
            {
                Dispose(true);
                GC.SuppressFinalize(this);
            }

            private void Dispose(bool disposing)
            {
                if (!disposed)
                {
                    font.Dispose();
                    format.Dispose();
                    gear.Dispose();
                    gearhole.Dispose();
                    
                    disposed = true;
                }
            }

            ~IntroRenderer()
            {
                Dispose(false);
            }
        }

        public EngineIntro()
        {
            updateables = new List<IntroState>(2);
            renderer = new IntroRenderer();
        }

        private void OnAnimationComplete(EventArgs e)
        {
            if (!eventRisen)
            {
                eventRisen = true;

                if (AnimationComplete != null)
                    AnimationComplete(this, e);

                GameEngine.PostMessage(new AnimationCompleteMessage());
            }
        }

        protected internal override IntroState CreateBuffer()
        {
            var updateable = new IntroState();
            updateables.Add(updateable);
            updateable.AnimationComplete += (sender, e) => OnAnimationComplete(e);
            return updateable;
        }

        protected internal override IStateRenderer<IntroState> Renderer
        {
            get { return renderer; }
        }

        private bool disposed;

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        private void Dispose(bool disposing)
        {
            if (!disposed)
            {
                foreach (IntroState updateable in updateables)
                    updateable.Dispose();
                renderer.Dispose();

                disposed = true;
            }
        }

        ~EngineIntro()
        {
            Dispose(false);
        }
    }

    public sealed class AnimationCompleteMessage : EngineMessage
    { }
}
